"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var lodash_1 = require("lodash");
var chalk_1 = require("chalk");
var log = function (str) { return console.error(str); };
var clear = function () {
    process.stdout.write('\x1Bc');
};
var Messages = /** @class */ (function () {
    function Messages() {
        var _this = this;
        this.done = function (stats) {
            clear();
            var hasErrors = stats.hasErrors();
            var hasWarnings = stats.hasWarnings();
            if (!hasErrors && !hasWarnings) {
                _this.success(stats);
                return;
            }
            if (hasErrors) {
                _this.errors(_this.extract(stats, 'errors'));
                return;
            }
            if (hasWarnings) {
                _this.warnings(_this.extract(stats, 'warnings'));
                return;
            }
        };
        this.invalid = function () {
            clear();
            log(chalk_1.default.gray('Compiling...'));
        };
        this.invalid = lodash_1.debounce(this.invalid, 100);
    }
    Messages.prototype.apply = function (compiler) {
        compiler.hooks.done.tap('messages', this.done);
        compiler.hooks.invalid.tap('messages', this.invalid);
    };
    Messages.prototype.success = function (stats) {
        var time = this.duration(stats);
        log(chalk_1.default.green('Compiled successfully in ' + time + 'ms'));
    };
    Messages.prototype.errors = function (errors) {
        var _this = this;
        var results = errors.map(function (error) { return _this.format(error.stack || error.message); });
        var syntaxErrors = results.filter(function (str) { return str.includes('Syntax error:'); });
        if (syntaxErrors.length) {
            results = syntaxErrors;
        }
        // First error is usually it; others usually the same
        var msg = results[0]
            .split('\n')
            .map(function (line, i) {
            if (i === 0)
                return chalk_1.default.inverse.red(line);
            return chalk_1.default.white(line);
        })
            .join('\n');
        log(msg);
    };
    Messages.prototype.warnings = function (warnings) {
        var _this = this;
        var results = warnings.map(function (warning) { return _this.format(warning.message); });
        results.map(function (result) { return log(chalk_1.default.yellow(result)); });
    };
    Messages.prototype.extract = function (stats, type) {
        var _this = this;
        if (stats.stats && stats.stats.length) {
            var errors = stats.stats.reduce(function (errors, stats) {
                return errors.concat(_this.extract(stats, type));
            }, []);
            // Dedupe to avoid showing the same error many times when multiple
            // compilers depend on the same module.
            return lodash_1.uniqBy(errors, function (error) { return error.message; });
        }
        return stats.compilation[type];
    };
    Messages.prototype.duration = function (stats) {
        var _this = this;
        if (stats.stats) {
            return stats.stats.reduce(function (time, stats) { return Math.max(time, _this.duration(stats)); }, 0);
        }
        return stats.endTime - stats.startTime;
    };
    Messages.prototype.format = function (message) {
        var lines = message.split('\n');
        if (lines.length > 2 && lines[1] === '') {
            lines.splice(1, 1); // Remove extra newline.
        }
        // Remove loader notation from filenames:
        //   `./~/css-loader!./src/App.css` ~~> `./src/App.css`
        if (lines[0].lastIndexOf('!') !== -1) {
            lines[0] = lines[0].substr(lines[0].lastIndexOf('!') + 1);
        }
        // Remove useless `entry` filename stack details
        lines = lines.filter(function (line) { return line.indexOf(' @ ') !== 0; });
        // 0 ~> filename; 1 ~> main err msg
        if (!lines[0] || !lines[1]) {
            return lines.join('\n');
        }
        // Cleans up verbose "module not found" messages for files and packages.
        if (lines[1].startsWith('Module not found: ')) {
            lines = [
                lines[0],
                lines[1] // "Module not found: " is enough detail
                    .replace("Cannot resolve 'file' or 'directory' ", '')
                    .replace('Cannot resolve module ', '')
                    .replace('Error: ', '')
                    .replace('[CaseSensitivePathsPlugin] ', '')
            ];
        }
        // Cleans up syntax error messages.
        if (lines[1].startsWith('Module build failed: ')) {
            lines[1] = lines[1].replace('Module build failed: SyntaxError:', 'Syntax error:');
        }
        var exportRegex = /\s*(.+?)\s*(")?export '(.+?)' was not found in '(.+?)'/;
        if (lines[1].match(exportRegex)) {
            lines[1] = lines[1].replace(exportRegex, "$1 '$4' does not contain an export named '$3'.");
        }
        // Reassemble & Strip internal tracing, except `webpack:` -- (create-react-app/pull/1050)
        var stackRegex = /^\s*at\s((?!webpack:).)*:\d+:\d+[\s\)]*(\n|$)/gm;
        return lines
            .join('\n')
            .replace(stackRegex, '')
            .trim();
    };
    return Messages;
}());
exports.default = Messages;
