import Config from '../server/config/webpack'
import * as Webpack from 'webpack'
import * as archy from 'archy'

export type AnalyzeOptions = {
  aliases?: { [alias: string]: string }
  environment?: 'development' | 'production'
  outpath?: string
  root?: string
}

type WebpackModule = {
  identifier: string
  issuer: string
  index: string | null
  id: string
  name: string
  size: number
  reasons: any
  source: string
}

type Module = {
  id?: string
  name: string
  size?: number
  reasons?: Object[]
  children: Module[]
}

export default async function Analyze(params: AnalyzeOptions) {
  const env = params.environment || 'production'
  const root = params.root || process.cwd()
  const aliases = params.aliases || {}

  // start loading the server-side config
  const s = Config({
    environment: env,
    target: 'server',
    action: 'analyze',
    routes: [],
    aliases,
    root
  })

  // start loading the client-side config
  const b = Config({
    environment: env,
    target: 'browser',
    action: 'analyze',
    routes: [],
    aliases,
    root
  })

  // build the configurations in parallel
  const stats = await Promise.all(
    [s, b].map(async config => {
      const c: Webpack.Configuration = await config
      const webpack = Webpack(c)
      return new Promise<Webpack.Stats>((resolve, reject) => {
        webpack.run((err: Error, stats: Webpack.Stats) => {
          if (err) return reject(err)

          if (stats.hasErrors()) {
            return reject(new Error(stats.toJson().errors.join('\n\n')))
          }

          return resolve(stats)
        })
      })
    })
  )

  const [, browserStats] = stats
  let browser = browserStats.toJson({ chunkModules: true })
  console.error(archy(formatTree(buildTree(browser.modules))))
}

function formatTree(tree: Module): archy.Data {
  const line: archy.Data = {
    label: tree.size ? `${tree.name}: ${formatSize(tree.size)}` : tree.name,
    nodes: []
  }

  if (!tree.children.length) return line
  line.nodes = tree.children.map(formatTree)
  return line
}

function buildTree(modules: WebpackModule[]): Module {
  const root: Module = {
    name: 'root',
    children: []
  }

  modules.forEach(module => {
    // remove this module if either:
    // - index is null
    // - issued by extract-text-plugin
    let extractInIdentifier =
      module.identifier.indexOf('extract-text-webpack-plugin') !== -1
    let extractInIssuer =
      module.issuer &&
      module.issuer.indexOf('extract-text-webpack-plugin') !== -1
    if (extractInIdentifier || extractInIssuer || module.index === null) {
      return
    }

    let mod: Module = {
      id: module.id,
      name: module.name,
      size: module.size,
      // TODO: show who imported it
      // reasons: module.reasons,
      children: []
    }

    let filename = module.name
    if (module.name.slice(0, 2) === './') {
      filename = filename.slice(2)
    }

    walk(mod, filename, root)
  })

  return root
}

function walk(module: Module, filename: string, parent: Module) {
  const idx = filename.indexOf('/')

  // leaf node
  if (!~filename.indexOf('/')) {
    module.name = filename
    parent.children.push(module)
    return
  }

  let folder = filename.slice(0, idx)
  if (folder === '~') {
    folder = 'node_modules'
  }

  let childFolder = getChild(parent.children, folder)
  if (!childFolder) {
    childFolder = {
      name: folder,
      children: []
    }
    parent.children.push(childFolder)
  }

  walk(module, filename.slice(idx + 1), childFolder)
}

function getChild(arr: Module[], name: string) {
  for (let i = 0; i < arr.length; i++) {
    if (arr[i].name === name) {
      return arr[i]
    }
  }
  return
}

function formatSize(size: number, precision = 1) {
  let kb = {
    label: 'k',
    value: 1024
  }
  let mb = {
    label: 'M',
    value: 1024 * 1024
  }
  let denominator

  if (size >= mb.value) {
    denominator = mb
  } else {
    denominator = kb
    if (size < kb.value * 0.92 && precision === 0) {
      precision = 1
    }
  }
  return (size / denominator.value).toFixed(precision) + denominator.label
}
