/*
 * Module dependencies
 */

import * as formatStats from 'webpack-format-messages'
import Config from '../server/config/webpack'
import Deferred from '../utils/deferred'
import Routes from '../server/routes'
import MemoryFS = require('memory-fs')
import * as Webpack from 'webpack'
import pall from '../utils/pall'
import { resolve } from 'path'
import { File } from '../elmo'

export type CompileOptions = {
  redirects?: { [module: string]: string }
  aliases?: { [module: string]: string }
  environment?: 'development' | 'production'
  root?: string
}

export default async function Compile(
  params: CompileOptions
): Promise<File[] | Error> {
  const root = params.root ? resolve(process.cwd(), params.root) : process.cwd()
  const env = params.environment || 'production'
  const aliases = params.aliases || {}
  const fs = new MemoryFS()

  // TODO: support compiling the redirects passed through
  const routes = await Routes(root)

  // start loading the server-side config
  const s = Config({
    environment: env,
    target: 'server',
    action: 'build',
    aliases,
    routes,
    root
  })

  // start loading the client-side config
  const b = Config({
    environment: env,
    target: 'browser',
    action: 'build',
    aliases,
    routes,
    root
  })

  const configs = await pall([s, b])
  if (configs instanceof Error) return configs

  // build the configurations in parallel
  const webpack = Webpack(configs)
  ;(webpack as any).outputFileSystem = fs

  const deferred = new Deferred<Webpack.Stats | Error>()

  webpack.run((err: any, stats: Webpack.Stats) => {
    // TODO: turn into a custom error
    if (err) {
      if (err instanceof Error) {
        return deferred.resolve(err)
      }

      // handle missing module and maybe more
      const nerr = new Error(err.message || err)
      nerr.stack = err.stack || ''
      return deferred.resolve(nerr)
    }

    // TODO: turn into a custom error
    const serr = checkError(stats)
    if (serr) {
      return deferred.resolve(serr)
    }

    return deferred.resolve(stats)
  })

  const stats = await deferred.wait()
  if (stats instanceof Error) return stats

  const files = statsToFiles(fs, stats)
  if (files instanceof Error) return files

  return files
}

function checkError(stats: Webpack.Stats): Error | null {
  const messages = formatStats(stats)

  if (messages.errors.length) {
    const errs = messages.errors.join('')
    return new Error(`webpack: failed to compile.\n\n${errs}`)
  }

  if (messages.warnings.length) {
    const warnings = messages.warnings.join('')
    return new Error(`webpack: compiled with warnings.\n\n${warnings}`)
  }

  return null
}

function statsToFiles(fs: MemoryFS, stats: Webpack.Stats): File[] | Error {
  const json = stats.toJson({ assets: true, chunks: true })
  if (!json) return new Error('webkpack: no json emitted')

  const files: File[] = []

  for (let i = 0; i < json.children.length; i++) {
    const child = json.children[i]
    const assets = child.assets

    if (!Array.isArray(assets)) continue

    for (let j = 0; j < assets.length; j++) {
      const asset = assets[j]

      const path = asset.name
      const contents = fs.readFileSync(resolve('/', path), 'utf8')

      files.push({ path, contents })
    }
  }

  return files
}
