/*
 * Module dependencies
 */

import * as WebpackDevMiddleware from 'webpack-dev-middleware'
import * as WebpackHotMiddleware from 'webpack-hot-middleware'
import Config from '../server/config/webpack'
import { resolve, extname } from 'path'
import Routes from '../server/routes'
import * as Webpack from 'webpack'
import * as express from 'express'
import pall from '../utils/pall'
import * as http from 'http'
import Messages from '../server/plugins/messages'

const isInteractive = !!process.stdout.isTTY

const regexps: { [s: string]: RegExp } = {}

export type ServeOptions = {
  redirects?: { [module: string]: string }
  aliases?: { [module: string]: string }
  environment?: 'development' | 'production'
  root?: string
  port?: number
}

export default async function serve(
  params: ServeOptions
): Promise<http.Server | Error> {
  const root = params.root ? resolve(process.cwd(), params.root) : process.cwd()
  const env = params.environment || 'development'
  const aliases = params.aliases || {}
  const app = express()

  // resolve routes
  const routes = await Routes(root)

  // start loading the server-side config
  const s = Config({
    environment: env,
    target: 'server',
    action: 'serve',
    routes: routes,
    aliases,
    root
  })

  // start loading the client-side config
  const b = Config({
    environment: env,
    target: 'browser',
    action: 'serve',
    routes: routes,
    aliases,
    root
  })

  const configs = await pall([s, b])
  if (configs instanceof Error) return configs

  // build the configurations in parallel
  const webpack = Webpack(configs)

  // apply the plugin to the multicompiler
  if (isInteractive) {
    // right now hooks is not on the @types/webpack multicompiler
    new Messages().apply(webpack as any)
  }

  // support custom routing
  if (routes.length) {
    // rehydrate regexps
    app.use(function(req, _res, next) {
      if (req.method !== 'GET') return next()
      if (extname(req.path) !== '') return next()

      for (let i = 0, len = routes.length; i < len; i++) {
        const route = routes[i]
        const regexp = (regexps[route.regexp] =
          regexps[route.regexp] || new RegExp(`${route.regexp}`))
        if (!regexp.test(req.path)) {
          continue
        }
        req.url = route.path.replace(/\/+$/, '') + '/'
        return next()
      }
      return next()
    })
  }

  const dev = WebpackDevMiddleware(webpack, {
    publicPath: '/',
    logLevel: 'silent'
  })

  app.use(dev)

  const hot = WebpackHotMiddleware(webpack as any, {
    log: false,
    path: '/__webpack_hmr',
    heartbeat: 10 * 1000
  })

  app.use(hot)

  // TODO: configurable
  app.use(express.static(resolve(root, 'static')))

  const server = http.createServer(app)
  const serverClose = server.close

  // also close the middleware
  server.close = (fn?: Function): http.Server => {
    return serverClose.call(server, fn)
  }

  return server
}
