import { Options, Props, Upgrader } from './render'
import { h } from 'preact'

// deeply convert an XML DOM to VDOM
export default function compile(
  n: Node,
  upgrader?: Upgrader,
  options: Options = { type: 'html' },
  index?: number,
  arr?: Node[]
): JSX.Element | string | null {
  // handle text nodes
  if (n.nodeType === 3) {
    return compileText(n, index, arr)
  }

  // handle everything else
  if (n.nodeType !== 1) return null

  const nodeName = String(n.nodeName).toLowerCase()

  // recursively compile the children
  const children = Array.prototype.map
    .call(n.childNodes, (child: Node, i: number, arr: Node[]) =>
      compile(child, upgrader, options, i, arr)
    )
    .filter((child: Node) => child)

  let component = h(
    nodeName,
    getProps((n as HTMLElement).attributes),
    children.length ? children : null
  )

  if (upgrader) return upgrade(component, upgrader)
  return component
}

// compile text nodes, trying to emulate HTML semantics by default, but these differ from JSX semantics
function compileText(n: Node, index?: number, arr: Node[] = []): string | null {
  let text = n.textContent !== null ? n.textContent : n.nodeValue || ''
  const isFirstOrLast = index === 0 || index === arr.length - 1

  // trim strings but don't entirely collapse whitespace
  if (text.match(/^[\s\n]+$/g)) {
    text = ' '
  }

  // skip leading/trailing whitespace
  if ((!text || text === ' ') && arr.length > 1 && isFirstOrLast) return null

  return text
}

function getProps(attrs: NamedNodeMap): Props {
  var len = attrs && attrs.length
  if (!len) return {}
  var props: Props = {}
  for (var i = 0; i < len; i++) {
    var name = attrs[i].name
    var value = attrs[i].value
    props[name] = value === '' ? true : value
  }
  return props
}

/**
 * Upgrade our component to a virtual DOM node
 */

function upgrade(
  component: JSX.Element,
  upgrader?: Upgrader
): JSX.Element | null {
  if (typeof component.nodeName === 'function') return component
  var name = component.nodeName.toLowerCase()

  if (upgrader && upgrader.hasOwnProperty(name)) {
    var props: Props = {}
    for (var k in component.attributes) {
      if (!component.attributes.hasOwnProperty(k)) continue
      props[k] = component.attributes[k]
    }
    props.key = component.key || undefined
    props.children = component.children

    var newNode = upgrader[name](props)
    if (!newNode) return null
    component.nodeName = newNode.nodeName
    component.attributes = newNode.attributes
    component.children = newNode.children
    component.key = newNode.key
  } else {
    component.nodeName = name.replace(/[^a-z0-9-]/i, '')
  }

  return component
}
