import * as htmlparser2 from 'htmlparser2'
import * as blocks from 'block-elements'
import * as inlines from 'inline-elements'
// import * as inv from 'invisibles'
import * as unified from 'unified'
import * as remarkParse from 'remark-parse'
import * as remark2rehype from 'remark-rehype'
import * as rehypeStringify from 'rehype-stringify'
// import * as toSlug from 'to-slug-case'
import * as domSerializer from 'dom-serializer'
// import * as preactHTML from 'preact-html'
// import * as preactRender from 'preact-render-to-string'
import * as prettier from 'prettier'
import * as buble from 'buble'
// import * as path from 'path'
// const cwd = process.cwd()

// const preact = require('preact')
// const result = HTML(`
// # hey ho

// > hi **world**

// <Link><a>**hi**</a></Link>

// <ImageRequest Cool />

// > blockquote

// \`<div>**hi**</div>\`

// \`\`\`html
// <Link>
//   <a>
//     **hi**
//   </a>
// </Link>
// \`\`\`

// * li
// * li
// * li
// `, {
//   'Link': require('./link'),
//   a: (props) => preact.h('a', { link: true }, props.children)
// })

export interface HTMLNode {
  type: string
  children?: HTMLNode[]
}

export interface HTMLText extends HTMLNode {
  type: 'text'
  value: string
}

export interface HTMLElement extends HTMLNode {
  type: 'element'
  tagName: string
  properties: { [attr: string]: string }
}

interface HP2Node extends HTMLNode {
  parent?: HP2Node
  next?: HP2Node
  prev?: HP2Node
}

interface HP2Text extends HP2Node {
  data: string
}

interface HP2Element extends HP2Node {
  name: string
  attribs: { [attr: string]: string }
}

export interface Options {
  resolve?: (node: HTMLElement) => string
}

// implementation
export function JS(code: string, options?: Options) {
  options = options || {}
  const resolve = options.resolve
  const ast = parse(code)

  const requires: [string, string][] = [['preact', 'preact']]
  walk(ast, node => {
    // all components must have path resolves
    if (resolve && isComponent(node)) {
      const p = resolve(node)
      if (!p) throw new Error(`Couldn't find '${node.tagName}'.`)
      requires.push([node.tagName, p])
      return
    }

    // optionally override regular HTML tags with paths
    if (resolve && isTag(node)) {
      const p = resolve(node)
      if (p) requires.push([node.tagName, p])
      return
    }
  })

  const html = renderHTML(ast)
  const js = buble.transform(html, { jsx: 'preact.h' }).code
  const imports = requires
    .map(req => `import ${req[0]} from "${req[1]}"`)
    .join('\n')

  const result = `
${imports}
export default function () { return (${js}) }
  `

  const formatted = prettier.format(result, { semi: false })
  return formatted
  // console.log(eval(result))
}

// function resolver(node: HTMLElement) {
// }

// function HTML(code: String, components) {
//   const ast = parse(code)
//   const html = renderHTML(ast)
//   const vnodes = preactHTML(html, components, { type: 'xml' })
//   return preactRender(vnodes[0])
// }

function parse(code: string) {
  // first render code blocks and inline
  // code because they may have HTML inside
  // of them.
  code = code.replace(/``?`?([^`]+)``?`?/g, escapeHtml)

  // get our HTML AST
  const hast = parseHTML(code)

  // walk our text nodes turning them into markdown
  walk(hast, (node, parent) => {
    if (!isText(node) || isCodeTag(parent)) return
    let text = node.value

    // ignore empty tags
    if ((text || '').trim() === '') return

    // parse the markdown
    const contents = processMarkdown(text)
    const subtree = parseHTML(contents)

    replace(node, subtree.children, parent)
  })

  // some sanitization and cleanup rules
  walk(hast, function(node, parent) {
    blockInsideInline(node, parent)
    collapseP(node, parent)
    collapsePInsideHeaders(node, parent)
  })

  return hast
}

function processMarkdown(code: string) {
  return unified()
    .use(remarkParse)
    .use(remark2rehype)
    .use(rehypeStringify)
    .processSync(code).contents
}

function parseHTML(code: string) {
  const ast = htmlparser2.parseDOM(`<div class="mdx">${code}</div>`, {
    xmlMode: true
  })[0]

  // turn the AST into a proper HAST
  walk(ast, function(node) {
    const n = node

    if (isHP2Node(node)) {
      delete node.next
      delete node.prev
      delete node.parent
    }

    // handle tags
    if (isHP2Tag(n) && isTag(node)) {
      node.type = 'element'
      node.tagName = n.name
      node.properties = n.attribs
    }

    // handle text
    if (isHP2Text(n) && isText(node)) {
      // htmlparser2 adds this entity onto our
      // escaped HTML, we remove it here.
      node.value = n.data.replace(/#x26;/g, '')
    }
  })

  return ast
}

function renderHTML(ast: HTMLNode) {
  // turn the HAST back
  walk(ast, function(node) {
    const n = node

    // handle tags
    if (isHP2Tag(node) && isTag(n)) {
      node.type = 'tag'
      node.name = n.tagName
      node.attribs = n.properties
    }

    // handle text
    if (isHP2Text(node) && isText(n)) {
      node.data = n.value
    }
  })

  return domSerializer(ast, { xmlMode: true })
}

function isCodeTag(node?: HTMLNode): node is HTMLElement {
  return isTag(node) && node.tagName === 'code'
}

function isPTag(node?: HTMLNode): node is HTMLElement {
  return isTag(node) && node.tagName === 'p'
}

function isHeaderTag(node?: HTMLNode): node is HTMLElement {
  return isTag(node) && /^h[1-6]$/.test(node.tagName)
}

function isComponent(node?: HTMLNode): node is HTMLElement {
  return isTag(node) && node.tagName[0] === node.tagName[0].toUpperCase()
}

function isTag(node?: HTMLNode): node is HTMLElement {
  return !!(node && (node.type === 'element' || node.type === 'tag'))
}

function isText(node?: HTMLNode): node is HTMLText {
  return !!(node && node.type === 'text')
}

function isHP2Node(node: HTMLNode): node is HP2Node {
  return (
    node.type === 'tag' || 'parent' in node || 'next' in node || 'prev' in node
  )
}

function isHP2Text(node: HP2Node): node is HP2Text {
  return 'data' in node
}

function isHP2Tag(node: HP2Node): node is HP2Element {
  return node && (node.type === 'element' || node.type === 'tag')
}

function walk(
  node: HTMLNode,
  fn: (node: HTMLNode, parent?: HTMLNode) => void,
  parent?: HTMLNode
) {
  const children = node.children || []
  for (let i = 0; i < children.length; i++) {
    walk(children[i], fn, node)
  }
  fn(node, parent)
}

function blockInsideInline(node: HTMLNode, parent?: HTMLNode) {
  if (isBlock(node) && isInline(parent)) {
    unwrap(node, parent)
  }
}

function collapseP(node: HTMLNode, parent?: HTMLNode) {
  if (isPTag(node) && isPTag(parent)) {
    unwrap(node, parent)
  }
}

function collapsePInsideHeaders(node: HTMLNode, parent?: HTMLNode) {
  if (isPTag(node) && isHeaderTag(parent)) {
    unwrap(node, parent)
  }
}

// function unescapeCode(node?: HTMLNode, parent?: HTMLNode) {
//   if (!isText(node) || !isCodeTag(parent)) return
//   console.log('beforez', node.value)
//   node.value = node.value
//     .replace(/&quot;/g, '"')
//     .replace(/&amp;/g, '&')
//     .replace(/&#39;/g, "'")
//     .replace(/&lt;/g, '<')
//     .replace(/&gt;/g, '>')
//   console.log('afterz', node.value)
// }

function replace(node: HTMLNode, replacements: HTMLNode[], parent?: HTMLNode) {
  if (!parent) return replacements
  const children = parent.children || []
  const i = children.indexOf(node)
  const newChildren = children
    .slice(0, i)
    .concat(replacements)
    .concat(children.slice(i + 1))
  parent.children = newChildren
  return replacements
}

function unwrap(node: HTMLNode, parent?: HTMLNode) {
  return replace(node, node.children || [], parent)
}

function isBlock(node?: HTMLNode) {
  return isTag(node) && !!~blocks.indexOf(node.tagName.toLowerCase())
}

function isInline(node?: HTMLNode): boolean {
  // console.log('inlines', inlines)
  return isTag(node) && !!~inlines.indexOf(node.tagName.toLowerCase())
}

function escapeHtml(string: string): string {
  var str = '' + string
  var match = /["'&<>]/.exec(str)

  if (!match) {
    return str
  }

  var escape
  var html = ''
  var index = 0
  var lastIndex = 0

  for (index = match.index; index < str.length; index++) {
    switch (str.charCodeAt(index)) {
      case 34: // "
        escape = '&quot;'
        break
      case 38: // &
        escape = '&amp;'
        break
      case 39: // '
        escape = '&#39;'
        break
      case 60: // <
        escape = '&lt;'
        break
      case 62: // >
        escape = '&gt;'
        break
      default:
        continue
    }

    if (lastIndex !== index) {
      html += str.substring(lastIndex, index)
    }

    lastIndex = index + 1
    html += escape
  }
  return lastIndex !== index ? html + str.substring(lastIndex, index) : html
}
