import * as ts from 'typescript'
import * as Stylis from 'stylis'

// setup stylis
const stylis = new Stylis()
stylis.set({
  cascade: false,
  compress: true
})

export interface Result extends ts.TranspileOutput {
  styles: {}[]
}

export type Options = {
  importName?: string
  importFrom?: string
  runtime?: boolean
  target?: 'server' | 'browser'
}

// This is a lot like a babel plugin, but using
// typescript's native compiler instead.

// There's a bunch of ceremony just to get the
// the transform setup, but this API would
// allow us to attach multiple transforms
// and only traverse the AST once.

export type Visitors<T> = {
  [type: number]:
    | ((node: T, context?: ts.TransformationContext) => T | void)
    | {
        enter?: (node: T, context?: ts.TransformationContext) => T | void
        exit?: (node: T, context?: ts.TransformationContext) => T | void
      }
}

export default function Transform<T extends ts.Node>(
  opts?: Options
): Visitors<T> {
  const options = opts || {}
  options.runtime =
    typeof options.runtime === 'undefined' ? false : options.runtime
  const importFrom = options.importFrom || 'style'
  options.importName = options.importName || 'StyledTSX'
  const importName = options.importName

  return {
    [ts.SyntaxKind.JsxElement]: {
      exit: node => {
        let el = (node as ts.Node) as ts.JsxElement
        const updates: [number, string, string][] = []
        const updated: Array<ts.JsxChild> = []
        const hashes: Array<string> = []
        let children = el.children

        for (let i = 0, len = children.length; i < len; i++) {
          if (!isStyledJSX(options, children[i])) {
            continue
          }

          // we know at this point its an element
          const child = children[i] as ts.JsxElement

          // ignore if no children
          const css = getCSS(child)
          if (!css) continue
          const id = 'tsx-' + hash(css)
          const compiled = stylis('.' + id, css)

          // store the updates until we
          // make changes to the parent
          // first
          updates.push([i, id, compiled])
          hashes.push(id)
        }

        // if this parent doesn't have a <style> tag, return
        if (!updates.length) return

        // update the class on the parent recursively
        const parent = updateClassRecursively(options, el, hashes.join(' '))
        // a parent can't really be self-closing
        // TODO: we should probably clean this up with a public and private
        // updateClassRecursively.
        if (!parent || parent.kind === ts.SyntaxKind.JsxSelfClosingElement)
          return

        // update children after we updated all the nodes
        children = parent.children

        // now that we've added classes, we're going to
        // swap the style tag out. this order maintains
        // the references.
        let update = updates.shift()
        for (let i = 0, len = children.length; i < len; i++) {
          if (!update || update[0] !== i) {
            updated.push(children[i])
            continue
          } else if (!options.runtime) {
            continue
          }

          const child = children[i] as ts.JsxElement

          // update the child
          const styledTSX = ts.updateJsxElement(
            child,
            ts.updateJsxOpeningElement(
              child.openingElement,
              ts.createIdentifier(importName),
              ts.createJsxAttributes([
                ts.createJsxAttribute(
                  ts.createIdentifier('id'),
                  ts.createLiteral(update[1])
                ),
                ts.createJsxAttribute(
                  ts.createIdentifier('css'),
                  ts.createLiteral(update[2])
                )
              ])
            ),
            [],
            ts.updateJsxClosingElement(
              child.closingElement,
              ts.createIdentifier(importName)
            )
          )

          updated.push(styledTSX)
          update = updates.shift()
        }

        // update the jsx element with the new children
        const replacement = ts.updateJsxElement(
          el,
          parent.openingElement,
          updated,
          parent.closingElement
        )

        return (replacement as ts.Node) as T
      }
    },
    [ts.SyntaxKind.SourceFile]: (node, context) => {
      if (!context) return
      const opts = context.getCompilerOptions()
      let { module } = opts

      if (module === ts.ModuleKind.CommonJS) {
        const file = (node as ts.Node) as ts.SourceFile
        const update = ts.updateSourceFileNode(file, [
          ts.createVariableStatement(
            /*modifiers*/ undefined,
            ts.createVariableDeclarationList([
              ts.createVariableDeclaration(
                importName,
                /*type*/ undefined,
                ts.createPropertyAccess(
                  ts.createCall(
                    ts.createIdentifier('require'),
                    [],
                    [ts.createLiteral(importFrom)]
                  ),
                  ts.createIdentifier('default')
                )
              )
            ])
          ),
          ...file.statements
        ])
        return (update as ts.Node) as T
      }

      // import StyledTSX from 'styled-tsx'
      if (module === ts.ModuleKind.ES2015 || module === ts.ModuleKind.ESNext) {
        const file = (node as ts.Node) as ts.SourceFile
        const update = ts.updateSourceFileNode(file, [
          ts.createImportDeclaration(
            /* decorators */ undefined,
            /* modifiers */ undefined,
            ts.createImportClause(ts.createIdentifier(importName), undefined),
            ts.createLiteral(importFrom)
          ),
          ...file.statements
        ])
        return (update as ts.Node) as T
      }

      return node
    }
  }
}

// function isScoped(
//   node: ts.JsxChild
// ): boolean {
//   // ignore non-jsx elements
//   if (node.kind !== ts.SyntaxKind.JsxElement) {
//     return false
//   }

//   const n = node.openingElement

//   // loop over properties
//   const props = n.attributes.properties || []

//   return true
// }

function isStyledJSX(
  options: Options,
  node: ts.JsxChild
): { global: boolean; scoped: boolean } | false {
  // ignore non-jsx elements
  if (node.kind !== ts.SyntaxKind.JsxElement) {
    return false
  }

  const n = node.openingElement
  if (n.tagName.kind !== ts.SyntaxKind.Identifier) {
    throw new Error(
      'Expected an identifier but instead got a node of kind: ' + n.tagName.kind
    )
  }
  const tag = n.tagName as ts.Identifier

  // if we've already renamed this tag, return a true
  // TODO: this is a hack
  if (tag.text === options.importName) {
    return { global: false, scoped: true }
  }

  // ignore any tags that aren't <style>
  if (tag.text !== 'style') return false

  // loop over properties
  const props = n.attributes.properties || []
  let isGlobal = false
  let isScoped = false
  for (let i = 0, len = props.length; i < len; i++) {
    const name = props[i].name
    if (!name) continue
    const key = getText(name)
    isScoped = key === 'jsx' || key === 'scoped'
    isGlobal = key === 'global'
  }

  // ignore if not: <style jsx>, <style global>, <style scoped>
  if (!isScoped && !isGlobal) return false

  return {
    global: isGlobal,
    scoped: isScoped
  }
}

function isHeadTag(node: ts.JsxChild): boolean {
  // jsx elements tags
  if (node.kind === ts.SyntaxKind.JsxElement) {
    const n = node.openingElement
    const nodeName = n.tagName.getText()
    if (nodeName !== 'Head') return false
    return true
  }

  // self-closing tags
  if (node.kind === ts.SyntaxKind.JsxSelfClosingElement) {
    const nodeName = node.tagName.getText()
    if (nodeName !== 'Head') return false
    return true
  }

  return false
}

export function updateClassRecursively(
  options: Options,
  node: ts.JsxElement | ts.JsxSelfClosingElement,
  hash: string
): ts.JsxElement | ts.JsxSelfClosingElement | void {
  // handle self-closing tags
  if (node.kind === ts.SyntaxKind.JsxSelfClosingElement) {
    return updateSelfClosingClass(node, hash)
  }

  const updated: Array<ts.JsxChild> = []
  const children = node.children

  // traverse down the children first
  for (let i = 0, len = children.length; i < len; i++) {
    const ch = children[i]

    if (
      ch.kind !== ts.SyntaxKind.JsxElement &&
      ch.kind !== ts.SyntaxKind.JsxSelfClosingElement
    ) {
      updated.push(ch)
      continue
    } else if (isStyledJSX(options, ch)) {
      // ignore <style scoped> tags
      updated.push(ch)
      continue
    } else if (isHeadTag(ch)) {
      // ignore <Head /> tags
      updated.push(ch)
      continue
    }

    const newChild = updateClassRecursively(options, ch, hash)
    updated.push(newChild || ch)
  }

  const opening = updateClass(node.openingElement, hash)
  if (!opening) return
  return ts.updateJsxElement(node, opening, updated, node.closingElement)
}

function updateSelfClosingClass(
  node: ts.JsxSelfClosingElement,
  hash: string
): ts.JsxSelfClosingElement | void {
  const attrs = node.attributes
  const props = attrs.properties || []

  for (let i = 0, len = props.length; i < len; i++) {
    // ignore spread attributes {...props}
    if (props[i].kind === ts.SyntaxKind.JsxSpreadAttribute) continue
    const prop = props[i] as ts.JsxAttribute

    // ignore expressions
    if (!prop.initializer) continue
    // if (prop.initializer.kind === ts.SyntaxKind.JsxExpression) continue

    const name = prop.name
    if (!name) continue

    const key = getText(name)
    if (key === 'class' || key === 'className') {
      let newAttr = null
      switch (prop.initializer.kind) {
        case ts.SyntaxKind.StringLiteral:
          newAttr = updateClassString(prop, hash)
          break
        case ts.SyntaxKind.JsxExpression:
          newAttr = updateClassExpression(prop, hash)
          break
      }

      if (!newAttr) continue

      // if (prop.initializer.kind === ts.SyntaxKind.StringLiteral) {
      //   newAttr = updateStringClass(prop, hash)
      //   // const classes = prop.initializer.text.split(/\s+/)
      //   // const newClass = ts.createLiteral(classes.concat(hash).join(' '))
      //   // const newAttr = ts.updateJsxAttribute(prop, name, newClass)
      // } else if ()

      // splice the prop
      let p = props
        .slice(0, i)
        .concat(newAttr)
        .concat(props.slice(i + 1))

      // update the attributes & exit early
      return ts.updateJsxSelfClosingElement(
        node,
        node.tagName,
        ts.updateJsxAttributes(node.attributes, p)
      )
    }
  }

  const newProps = props.concat(
    ts.createJsxAttribute(
      ts.createIdentifier('className'),
      ts.createLiteral(hash)
    )
  )

  return ts.updateJsxSelfClosingElement(
    node,
    node.tagName,
    ts.updateJsxAttributes(node.attributes, newProps)
  )
}

export function updateClass(
  node: ts.JsxOpeningElement,
  hash: string
): ts.JsxOpeningElement | void {
  const attrs = node.attributes
  const props = attrs.properties || []

  for (let i = 0, len = props.length; i < len; i++) {
    // ignore spread attributes {...props}
    if (props[i].kind === ts.SyntaxKind.JsxSpreadAttribute) continue
    const prop = props[i] as ts.JsxAttribute

    // ignore expressions
    if (!prop.initializer) continue
    // if (prop.initializer.kind === ts.SyntaxKind.JsxExpression) continue

    const name = prop.name
    if (!name) continue

    const key = getText(name)
    if (key === 'class' || key === 'className') {
      let newAttr = null
      switch (prop.initializer.kind) {
        case ts.SyntaxKind.StringLiteral:
          newAttr = updateClassString(prop, hash)
          break
        case ts.SyntaxKind.JsxExpression:
          newAttr = updateClassExpression(prop, hash)
          break
      }

      if (!newAttr) continue

      // if (prop.initializer.kind === ts.SyntaxKind.StringLiteral) {
      //   newAttr = updateStringClass(prop, hash)
      //   // const classes = prop.initializer.text.split(/\s+/)
      //   // const newClass = ts.createLiteral(classes.concat(hash).join(' '))
      //   // const newAttr = ts.updateJsxAttribute(prop, name, newClass)
      // } else if ()

      // splice the prop
      let p = props
        .slice(0, i)
        .concat(newAttr)
        .concat(props.slice(i + 1))

      // update the attributes & exit early
      return ts.updateJsxOpeningElement(
        node,
        node.tagName,
        ts.updateJsxAttributes(node.attributes, p)
      )
    }
  }

  const newProps = props.concat(
    ts.createJsxAttribute(
      ts.createIdentifier('className'),
      ts.createLiteral(hash)
    )
  )

  return ts.updateJsxOpeningElement(
    node,
    node.tagName,
    ts.updateJsxAttributes(node.attributes, newProps)
  )
}

function updateClassString(
  node: ts.JsxAttribute,
  hash: string
): ts.JsxAttribute | void {
  const initializer = node.initializer as ts.StringLiteral
  const existing = initializer.text.split(/\s+/)
  for (let i = 0, len = existing.length; i < len; i++) {
    if (existing[i].slice(0, 4) === 'tsx-') return node
  }
  const text = existing.concat(hash).join(' ')
  const newClass = ts.createLiteral(text)
  return ts.updateJsxAttribute(node, node.name, newClass)
}

function updateClassExpression(
  node: ts.JsxAttribute,
  hash: string
): ts.JsxAttribute | void {
  const initializer = node.initializer as ts.JsxExpression
  if (!initializer.expression) {
    return ts.createJsxAttribute(
      node.name,
      ts.createJsxExpression(undefined, ts.createLiteral(hash))
    )
  }

  let expr = initializer.expression
  switch (expr.kind) {
    // {"hi"} {""} {''}
    case ts.SyntaxKind.StringLiteral:
      return ts.createJsxAttribute(
        node.name,
        ts.createJsxExpression(
          undefined,
          updateStringLiteral(expr as ts.StringLiteral, hash) || expr
        )
      )

    // {`hi ${world}`}
    case ts.SyntaxKind.TemplateExpression:
      return ts.createJsxAttribute(
        node.name,
        ts.createJsxExpression(
          undefined,
          updateTemplateExpression(expr as ts.TemplateExpression, hash) || expr
        )
      )

    // {`hi`}
    case ts.SyntaxKind.NoSubstitutionTemplateLiteral:
      return ts.createJsxAttribute(
        node.name,
        ts.createJsxExpression(
          undefined,
          updateNoSubstitutionExpression(
            expr as ts.NoSubstitutionTemplateLiteral,
            hash
          ) || expr
        )
      )
    default:
      return ts.createJsxAttribute(
        node.name,
        ts.createJsxExpression(
          undefined,
          wrapExpression(expr as ts.Expression, hash) || expr
        )
      )
  }
}

function updateStringLiteral(
  node: ts.StringLiteral,
  hash: string
): ts.StringLiteral | void {
  const text = (node.text + ' ' + hash).trim()
  return ts.createLiteral(text)
}

function updateNoSubstitutionExpression(
  node: ts.NoSubstitutionTemplateLiteral,
  hash: string
): ts.NoSubstitutionTemplateLiteral | void {
  node.text = (node.text + ' ' + hash).trim()
  return node
}

function updateTemplateExpression(
  node: ts.TemplateExpression,
  hash: string
): ts.TemplateExpression | void {
  if (!node.templateSpans.length) return
  const lastSpan = node.templateSpans[node.templateSpans.length - 1]
  if (lastSpan.literal.kind !== ts.SyntaxKind.TemplateTail) return

  // TODO: can't figure out how to update the ts.TemplateTail
  // in a non-destructive way
  const tail = lastSpan.literal
  tail.text = tail.text + ' ' + hash

  return node
}

function wrapExpression(
  node: ts.Expression,
  hash: string
): ts.BinaryExpression | void {
  return ts.createBinary(
    ts.createParen(node),
    ts.SyntaxKind.PlusToken,
    ts.createLiteral(' ' + hash)
  )
}

function getCSS(el: ts.JsxElement): string {
  if (!el.children.length) return ''

  const css = el.children[0]

  switch (css.kind) {
    // just a text element (probably single line)
    case ts.SyntaxKind.JsxText:
      return (css as ts.JsxText).getText().trim()
    case ts.SyntaxKind.JsxExpression:
      const expr = (css as ts.JsxExpression).expression
      if (!expr) return ''
      if (expr.kind !== ts.SyntaxKind.NoSubstitutionTemplateLiteral) {
        throw new Error('<style scoped> does not support templates ${...}')
      }
      const literal = expr as ts.NoSubstitutionTemplateLiteral
      return literal.text || ''
  }

  return ''
}

function getText(node: ts.Node): string {
  switch (node.kind) {
    case ts.SyntaxKind.NoSubstitutionTemplateLiteral:
      return (node as ts.NoSubstitutionTemplateLiteral).text
    case ts.SyntaxKind.Identifier:
      return (node as ts.Identifier).text
    case ts.SyntaxKind.StringLiteral:
      return (node as ts.StringLiteral).text
    case ts.SyntaxKind.NumericLiteral:
      return (node as ts.NumericLiteral).text
    default:
      return ''
  }
}

/**
 * JS Implementation of MurmurHash2
 *
 * @author <a href="mailto:gary.court@gmail.com">Gary Court</a>
 * @see http://github.com/garycourt/murmurhash-js
 * @author <a href="mailto:aappleby@gmail.com">Austin Appleby</a>
 * @see http://sites.google.com/site/murmurhash/
 *
 * @param {string} str ASCII only
 * @return {string} Base 36 encoded hash result
 */
function hash(str: string): string {
  var l = str.length
  var h = l
  var i = 0
  var k

  while (l >= 4) {
    k =
      (str.charCodeAt(i) & 0xff) |
      ((str.charCodeAt(++i) & 0xff) << 8) |
      ((str.charCodeAt(++i) & 0xff) << 16) |
      ((str.charCodeAt(++i) & 0xff) << 24)

    k = (k & 0xffff) * 0x5bd1e995 + ((((k >>> 16) * 0x5bd1e995) & 0xffff) << 16)
    k ^= k >>> 24
    k = (k & 0xffff) * 0x5bd1e995 + ((((k >>> 16) * 0x5bd1e995) & 0xffff) << 16)

    h =
      ((h & 0xffff) * 0x5bd1e995 +
        ((((h >>> 16) * 0x5bd1e995) & 0xffff) << 16)) ^
      k

    l -= 4
    ++i
  }

  switch (l) {
    case 3:
      h ^= (str.charCodeAt(i + 2) & 0xff) << 16
      break
    case 2:
      h ^= (str.charCodeAt(i + 1) & 0xff) << 8
      break
    case 1:
      h ^= str.charCodeAt(i) & 0xff
      h =
        (h & 0xffff) * 0x5bd1e995 + ((((h >>> 16) * 0x5bd1e995) & 0xffff) << 16)
      break
  }

  h ^= h >>> 13
  h = (h & 0xffff) * 0x5bd1e995 + ((((h >>> 16) * 0x5bd1e995) & 0xffff) << 16)
  h ^= h >>> 15

  return (h >>> 0).toString(36)
}
