import Route from '../../lib/route'
import { loader } from 'webpack'

export default function(
  this: loader.LoaderContext,
  source: string,
  _sourceMap: string
) {
  this.cacheable()

  const { resourcePath, query: { root, routes } } = this
  const route = getRoute(root, resourcePath)
  if (!route) {
    return this.callback(new Error("couldn't find route for page"))
  }

  // This is a bit confusing because we only want to update once, but
  // need to satisfy 3 update conditions: 1st page load, HMR update,
  // and Router.push(...). We only want to update once when we load
  // the page and when there's an HMR update, but not when we async
  // re-route.
  //
  //   -> only do HMR in development mode
  //   if (module.hot) {
  //   -> accept the hot module, this is a noop when we're booting
  //     module.hot.accept()
  //
  //     -> if we're booting, this will be "idle", "apply" occurs
  //     -> when we've made a change, this will allow us to only
  //     -> target HMR updates
  //     if (module.hot.status() === 'apply') {
  //       const Router = require('elmo/router').default
  //       Router.update(Component, route).catch(console.error)
  //       return
  //     }
  //   }
  //
  //   -> This will not be triggered when we're using Router.push(...)
  //   -> there is no corresponding <script>.
  //   if (document.querySelector('script[src="${route}"]')) {
  //     const Router = require('elmo/router').default
  //     Router.update(Component, route).catch(console.error)
  //   }
  const newSource = `
  ${source}
  ;(function (Component, route, routes) {
    Component = Component.default || Component
    const Router = require('elmo').Router
    const router = new Router(routes)

    if (module.hot) {
      module.hot.accept()

      // this should only occur when we are using
      // HMR, never when we boot or when we use Router.push(...)
      if (module.hot.status() === 'apply') {
        router.update(Component, route).catch(console.error)
        return
      }
    }

    // this should only get rendered on boot, never
    // during HMR or when we use Router.push(...)
    if (document.querySelector('script[src="'+ route +'"]')) {
      router.update(Component, route)
        .then(function () {
          // this will signal to prerender.io compatible
          // clients that the page has rendered successfully
          window.prerenderReady = true
        })
        .catch(console.error)
    }
  })(
    require('${this.resourcePath}'),
    ${JSON.stringify(route)},
    ${JSON.stringify(routes)}
  )
  `
  // console.log(newSource)
  this.callback(null, newSource, undefined)
}

function getRoute(root: string, path: string): string {
  const dir: string | void = [root].find(d => path.indexOf(d) === 0)
  if (!dir) return ''
  const route = Route(dir, path)
  if (route === '/') return '/index.js'
  return route + '/index.js'
}
