import { loader } from 'webpack'
import * as ts from 'typescript'
import Transform from '../../lib/style/transform'
import Typescript from '../../utils/typescript'
// import { validate } from '../config/tsconfig'

export default function(
  this: loader.LoaderContext,
  source: string,
  _sourceMap: string
) {
  const { resourcePath: path, query: { root, environment, target } } = this
  // mark loader as cacheable
  this.cacheable()

  // setup styled-tsx
  const hasRuntime = environment !== 'production' || target !== 'browser'
  const transformers = {
    before: [
      visitor(
        Transform({
          runtime: hasRuntime,
          target: target
        })
      )
    ]
  }

  // transpile
  const result = Typescript({ root, path, source, transformers, target })
  if (result instanceof Error) return this.callback(result)

  return this.callback(null, result.outputText)
}

type Visitors<T> = {
  [type: number]:
    | Visitor<T>
    | {
        enter?: Visitor<T>
        exit?: Visitor<T>
      }
}

type Visitor<T> = (node: T, context?: ts.TransformationContext) => T | void

function visitor(transform: Visitors<ts.Node>) {
  return (context: ts.TransformationContext) => {
    return (sourceFile: ts.SourceFile): ts.SourceFile => {
      return visit(sourceFile)
      function visit<T extends ts.Node>(node: T): T {
        const t = transform[node.kind]
        if (!t) {
          return ts.visitEachChild(node, visit, context)
        }

        // enter a node
        if (typeof t === 'object' && t.enter) {
          const tr = t.enter(node, context)
          if (tr) node = tr as T
        }

        // if it's a function, treat it
        // the same way as t.enter
        if (typeof t === 'function') {
          const tr = t(node, context)
          if (tr) node = tr as T
        }

        // visit each child
        let exitNode = ts.visitEachChild(node, visit, context)

        // if we have an exit call function
        if (typeof t === 'object' && t.exit) {
          const tr = t.exit(exitNode, context)
          if (tr) exitNode = tr as T
        }

        return exitNode
      }
    }
  }
}
