import { Plugin, Compiler, Stats } from 'webpack'
import { debounce, uniqBy } from 'lodash'
import chalk from 'chalk'

const log = (str: any) => console.error(str)
const clear = () => {
  process.stdout.write('\x1Bc')
}

export default class Messages implements Plugin {
  constructor() {
    this.invalid = debounce(this.invalid, 100)
  }

  apply(compiler: Compiler): void {
    compiler.hooks.done.tap('messages', this.done)
    compiler.hooks.invalid.tap('messages', this.invalid)
  }

  private done = (stats: Stats) => {
    clear()

    const hasErrors = stats.hasErrors()
    const hasWarnings = stats.hasWarnings()

    if (!hasErrors && !hasWarnings) {
      this.success(stats)
      return
    }

    if (hasErrors) {
      this.errors(this.extract(stats, 'errors'))
      return
    }

    if (hasWarnings) {
      this.warnings(this.extract(stats, 'warnings'))
      return
    }
  }

  private invalid = () => {
    clear()
    log(chalk.gray('Compiling...'))
  }

  private success(stats: Stats) {
    const time = this.duration(stats)
    log(chalk.green('Compiled successfully in ' + time + 'ms'))
  }

  private errors(errors: Error[]) {
    let results = errors.map(error => this.format(error.stack || error.message))
    const syntaxErrors = results.filter(str => str.includes('Syntax error:'))
    if (syntaxErrors.length) {
      results = syntaxErrors
    }

    // First error is usually it; others usually the same
    const msg = results[0]
      .split('\n')
      .map((line, i) => {
        if (i === 0) return chalk.inverse.red(line)
        return chalk.white(line)
      })
      .join('\n')

    log(msg)
  }

  private warnings(warnings: Error[]) {
    let results = warnings.map(warning => this.format(warning.message))
    results.map(result => log(chalk.yellow(result)))
  }

  private extract(stats: any, type: 'errors' | 'warnings'): Error[] {
    if (stats.stats && stats.stats.length) {
      const errors: Error[] = stats.stats.reduce(
        (errors: Error[], stats: any) =>
          errors.concat(this.extract(stats, type)),
        []
      )
      // Dedupe to avoid showing the same error many times when multiple
      // compilers depend on the same module.
      return uniqBy(errors, error => error.message)
    }

    return stats.compilation[type]
  }

  private duration(stats: any): number {
    if (stats.stats) {
      return stats.stats.reduce(
        (time: number, stats: any) => Math.max(time, this.duration(stats)),
        0
      )
    }
    return stats.endTime - stats.startTime
  }

  private format(message: string) {
    let lines = message.split('\n')

    if (lines.length > 2 && lines[1] === '') {
      lines.splice(1, 1) // Remove extra newline.
    }

    // Remove loader notation from filenames:
    //   `./~/css-loader!./src/App.css` ~~> `./src/App.css`
    if (lines[0].lastIndexOf('!') !== -1) {
      lines[0] = lines[0].substr(lines[0].lastIndexOf('!') + 1)
    }

    // Remove useless `entry` filename stack details
    lines = lines.filter(line => line.indexOf(' @ ') !== 0)

    // 0 ~> filename; 1 ~> main err msg
    if (!lines[0] || !lines[1]) {
      return lines.join('\n')
    }

    // Cleans up verbose "module not found" messages for files and packages.
    if (lines[1].startsWith('Module not found: ')) {
      lines = [
        lines[0],
        lines[1] // "Module not found: " is enough detail
          .replace("Cannot resolve 'file' or 'directory' ", '')
          .replace('Cannot resolve module ', '')
          .replace('Error: ', '')
          .replace('[CaseSensitivePathsPlugin] ', '')
      ]
    }

    // Cleans up syntax error messages.
    if (lines[1].startsWith('Module build failed: ')) {
      lines[1] = lines[1].replace(
        'Module build failed: SyntaxError:',
        'Syntax error:'
      )
    }

    const exportRegex = /\s*(.+?)\s*(")?export '(.+?)' was not found in '(.+?)'/
    if (lines[1].match(exportRegex)) {
      lines[1] = lines[1].replace(
        exportRegex,
        "$1 '$4' does not contain an export named '$3'."
      )
    }

    // Reassemble & Strip internal tracing, except `webpack:` -- (create-react-app/pull/1050)
    const stackRegex = /^\s*at\s((?!webpack:).)*:\d+:\d+[\s\)]*(\n|$)/gm
    return lines
      .join('\n')
      .replace(stackRegex, '')
      .trim()
  }
}
