import { Route } from '../../elmo'
// import rewrite from '../../utils/rewrite-route'
import { h } from 'preact'
import * as render_to_string from 'preact-render-to-string'
import { Plugin, Compiler, compilation } from 'webpack'
import evaluate from '../../utils/evaluate'
import { join, extname, basename, dirname } from 'path'
import flush from '../../lib/style/server'

export type Entries = {
  [s: string]: {
    source: () => string
    size: () => number
  }
}

export type EntryPaths = {
  [entry: string]: string
}

export type StaticParams = {
  environment: string
  elmoRoot: string
  target: string
  action: string
  root: string
  routes: Route[]
  port?: number
}

/**
 * Module Exports
 */

export default class Static implements Plugin {
  private readonly options: StaticParams

  constructor(options: StaticParams) {
    this.options = options
  }

  async build(entries: Entries): Promise<void | Error> {
    const { elmoRoot } = this.options

    // find the document
    const documentPath = Object.keys(entries).find(entry => {
      return basename(entry, extname(entry)) === '_document'
    })
    if (!documentPath) {
      return new Error("couldn't find the _document in the entries")
    }

    // turn the source into a function
    const Document = evaluate(documentPath, entries[documentPath].source(), {
      require
    })
    if (Document instanceof Error) {
      return Document
    } else if (typeof Document !== 'function') {
      return new Error("document doesn't have a default export")
    }

    // delete the document from the entries
    delete entries[documentPath]

    // turn the node files into html files
    //
    // TODO: try workerize-ing this
    for (let entry in entries) {
      // ignore non-HTML files
      if (extname(entry) !== '.html') continue

      const Page = evaluate(entry, entries[entry].source(), { require })
      if (Page instanceof Error) {
        return Page
      } else if (typeof Page !== 'function') {
        // TODO: map back to the original file
        return new Error(`"${entry}" doesn't have a default export`)
      }

      const common = h('script', {
        type: 'text/javascript',
        src: '/common.js'
      })

      const script = h('script', {
        type: 'text/javascript',
        src: '/' + join(dirname(entry), 'index.js')
      })

      const props = {} // TODO: support custom builds

      // we render here, so we can rewind
      const page = render_to_string(h(Page, props))

      const { head, style } = rewindAll(elmoRoot)

      // render the document into a string
      const html: string = render_to_string(
        h(Document, {
          page,
          common,
          script,
          head,
          style,
          props
        })
      )

      // update the entry
      entries[entry] = {
        source: () => html,
        size: () => html.length
      }
    }
  }

  apply(compiler: Compiler) {
    const self = this

    compiler.hooks.thisCompilation.tap(
      'static',
      (compilation: compilation.Compilation) => {
        compilation.hooks.optimizeAssets.tapAsync('static', function(
          assets: compilation.Asset[],
          done: (err?: Error) => void
        ) {
          self
            .build(assets as any)
            .then(e => {
              if (e) compilation.errors.push(e)
              done()
            })
            .catch(e => {
              done(e)
            })
        })
      }
    )
  }
}

/**
 * Get the initial properties of the static build
 * // TODO: better inputs
 */
// function getBuildProps(Component: any): Promise<Object> {
//   if (Component.getBuildProps) {
//     return Component.getBuildProps({ build: true, path: '', query: {} })
//   } else {
//     return Promise.resolve({})
//   }
// }

/**
 * Render the HTML
 */

// TODO: not exactly sure how to type the Document when there's an evaluate()
// async function renderHTML(
//   Document: any,
//   params: {
//     elmoRoot: string
//     filepath: string
//     Component: any
//     props: any
//   }
// ) {
//   const Component = params.Component
//   const page = Stringify(<Component {...params.props} />)
//   const { head, style } = rewindAll(params.elmoRoot)

//   const docProps = await getBuildProps(Document)

//   // create a virtual DOM tree
//   const vhtml = (
//     <Document
//       page={page}
//       head={head}
//       style={style}
//       props={docProps || {}}
//       filepath={params.filepath}
//     />
//   )

//   const html = '<!DOCTYPE html>' + Stringify(vhtml)

//   // ignore any <style scoped> or <Head /> in the document
//   rewindAll(params.elmoRoot)

//   return html
// }

/**
 * Add the entry file
 */

// function addFile(entries: Entries, entry: string, html: string) {
//   const name = basename(dirname(entry))
//   entries[name + '/index.html'] = {
//     source: () => html,
//     size: () => html.length
//   }
//   return null
// }

// Okay so the reason for this is that when we call evaluate
// it may use a non-cached version of <Head /> or <Style/>
// depending on which module we require. These need to be
// pull from the cache to maintain their flush state.

// This function normalizes over all possible places
// you can get these modules from, so no matter where
// we require it we'll get the state.
function rewindAll(
  elmoRoot: string
): { head: JSX.Element[]; style: JSX.Element[] } {
  const dist = join(elmoRoot, 'dist')
  const src = join(elmoRoot, 'src')
  const cache = require.cache

  const distHead = join(dist, 'lib', 'head.js')
  const srcHead = join(src, 'lib', 'head.ts')
  const distStyle = join(dist, 'lib', 'style', 'style.js')
  const srcStyle = join(src, 'lib', 'style', 'style.ts')
  let fn = null

  let head: JSX.Element[] = []
  let style: JSX.Element[] = []

  for (let req in cache) {
    if (!cache[req]) continue
    switch (req) {
      case distHead:
        fn = cache[req].exports.default
        head = head.concat(fn.rewind())
        break
      case srcHead:
        fn = cache[req].exports.default
        head = head.concat(fn.rewind())
        break
      case distStyle:
        fn = cache[req].exports.default
        style = style.concat(flush(fn.flush()))
        break
      case srcStyle:
        fn = cache[req].exports.default
        style = style.concat(flush(fn.flush()))
        break
    }
  }

  return { head, style }
}
