import { join, relative, extname } from 'path'
// import * as loaderUtils from 'loader-utils'
import { BaseError } from 'make-error'
import * as ts from 'typescript'
import json5 from './json5'

// based on ts-node's: https://goo.gl/V7Qpnp
const defaultCompilerOptions = {
  sourceMap: true,
  inlineSourceMap: false,
  inlineSources: true,
  declaration: false,
  noEmit: false
}

/**
 * Typescript interface
 */

export namespace Typescript {
  export interface Input {
    root: string
    path: string
    source: string
    target: 'browser' | 'server'
    transformers?: ts.CustomTransformers
  }
}

/**
 * Transpile function
 */

export default function transpile(
  input: Typescript.Input
): ts.TranspileOutput | Error {
  const { root, path, source, target, transformers } = input

  // read the tsconfig via json5
  const tsconfig = json5(join(root, 'tsconfig.json'))

  // use provided config or ts-node's defaults
  let compilerOptions = tsconfig.compilerOptions
    ? tsconfig.compilerOptions || {}
    : defaultCompilerOptions

  // default module support
  compilerOptions.module =
    target === 'server' ? ts.ModuleKind.CommonJS : ts.ModuleKind.ESNext

  // add in jsx compilation when the filepath is tsx
  if (extname(path) === '.tsx') {
    compilerOptions = Object.assign(compilerOptions, {
      jsx: 'react',
      jsxFactory: 'h'
    })
  }

  const result = ts.transpileModule(source, {
    fileName: path,
    reportDiagnostics: true,
    compilerOptions: compilerOptions,
    transformers: transformers || {}
  })

  // parse an error if there is one
  if (result.diagnostics && result.diagnostics.length) {
    return new TSError(...formatDiagnostics(result.diagnostics, root, 0))
  }

  return result
}

/**
 * Diagnostic interface
 */

interface TSDiagnostic {
  message: string
  code: number
}

/**
 * TypeScript diagnostics error.
 */
class TSError extends BaseError {
  name = 'TSError'

  constructor(...diagnostics: TSDiagnostic[]) {
    super(
      `⨯ Unable to compile TypeScript\n${diagnostics
        .map(x => x.message)
        .join('\n')}`
    )
  }
}

function formatDiagnostics(
  diagnostics: ts.Diagnostic[] = [],
  root: string,
  lineOffset: number
): TSDiagnostic[] {
  return diagnostics.map(x => formatDiagnostic(x, root, lineOffset))
}

// create an error if it exists
function formatDiagnostic(
  diagnostic: ts.Diagnostic,
  root: string,
  lineOffset: number
): TSDiagnostic {
  const messageText = ts.flattenDiagnosticMessageText(
    diagnostic.messageText,
    '\n'
  )

  const { code } = diagnostic
  if (!diagnostic.file) {
    return { message: `${messageText} (${code})`, code }
  }

  const path = relative(root, diagnostic.file.fileName)
  if (!diagnostic.start) {
    return { message: `${path}: ${messageText} (${code})`, code }
  }

  const { line, character } = diagnostic.file.getLineAndCharacterOfPosition(
    diagnostic.start
  )
  const message = `${path} (${line + 1 + lineOffset},${character +
    1}): ${messageText} (${code})`

  return { message, code }
}
